import sys
import locale as lo
import os
import re
import math
import argparse
import datetime, time
import xml.etree.cElementTree as ET
import subprocess
from collections import OrderedDict 
def is_valid_par_file(file_argument):
    """
    Checks if --INP argument points to a file with .par extension
    """
    file_ext = os.path.splitext(file_argument)[1]
    if not os.path.exists(file_argument):
        sys.exit("ERROR: mausbpf2eaf : The file %s does not exist! Please specify a valid BAS '.par' file as input - exiting" % file_argument)
    elif file_ext == '.par' or file_ext == '.PAR' or file_ext == '.bpf' or file_ext == '.BPF':
        return file_argument
    else:
        sys.exit("ERROR: mausbpf2eaf : The extension %s is not a valid extension! Please specify a valid BAS '.par' file as input - exiting" % file_ext)

def is_valid_output_file(file_path):
    """
    #Checks if --OUT argument is valid
    """
    file_ext = os.path.splitext(file_path)[1]
    if file_ext != '.eaf':
        sys.exit("ERROR: mausbpf2eaf : The extension '%s' is not a valid output extension. Please specify a '.eaf' output file - exiting" %file_ext)
    return True
    


def write_elan_data(data, out_file, orig_wav_path, verbosity_level):
    if verbosity_level >= 1:
        print("DEBUG: mausbpf2eaf : Constructing ELAN XML Data")
    if verbosity_level >= 2:
        print("DEBUG: mausbpf2eaf : Constructing .eaf header information")
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Defining XML Namespace and creating annotation document")
    NS = "http://www.w3.org/2001/XMLSchema-instance"
    location_attribute = '{%s}noNamespaceSchemaLocation' % NS
    annot_doc = ET.Element("ANNOTATION_DOCUMENT", attrib={location_attribute: "http://www.mpi.nl/tools/elan/EAFv3.0.xsd"})
    
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Calculating ISO timezone offset and setting DATE information.")
    #Calculate ISO offset and set date to utc with ISO offset in correct format
    utc_offset_sec = time.altzone if time.localtime().tm_isdst else time.timezone
    utc_offset = datetime.timedelta(seconds=-utc_offset_sec)
    annot_doc.set("DATE", datetime.datetime.utcnow().replace(microsecond=0).replace(tzinfo=datetime.timezone(offset=utc_offset)).isoformat())
    
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Setting AUTHOR as MAUS")
    annot_doc.set("AUTHOR", "MAUS")
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Setting VERSION to 3.0")
    annot_doc.set("VERSION", "3.0")
    
    header = ET.SubElement(annot_doc, "HEADER")
    header.set("TIME_UNITS", "milliseconds")
   
    #Media Descriptor 
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Defining reference audio file to be at '%s' and having Mime_type: 'audio/x-wav" % os.path.abspath(orig_wav_path))
    media_desc = ET.SubElement(header, "MEDIA_DESCRIPTOR")
    media_desc.set("MEDIA_URL", "file://" + os.path.abspath(orig_wav_path)) 
    
    """
    #Check for different MIME Types(Flac, MP3)
    NOT YET IMPLEMENTED IN ELAN (SEE  https://tla.mpi.nl/wp-content/uploads/2016/12/Video_encoding_guidelines_ELAN.pdf)
    
    if os.path.isfile(os.path.abspath(orig_wav_path)):
        mime = magic.Magic(mime=True)
        media_mime_type = mime.from_file(os.path.abspath(orig_wav_path))
        
        media_desc.set("MIME_TYPE", media_mime_type)
    else:
        media_desc.set("MIME_TYPE", "audio/x-wav")
        sys.stderr.write("WARNING: mausbpf2eaf : There is no audio file at '%s' from which to infer the mime type, assuming 'audio/x-wav'.\n"% os.path.abspath(orig_wav_path))
    """
    media_desc.set("MIME_TYPE", "audio/x-wav")
    media_desc.set("RELATIVE_MEDIA_URL", "./" + orig_wav_path)
    
    #Property (Contains lastAnnotationId = maxCount of Annotation elments)
    prop = ET.SubElement(header, "PROPERTY")
    
    #Make the time order
    make_time_order(data, annot_doc, verbosity_level)
    
    
    max_annot_idx = 0
    max_annot_idx = make_tiers(data, annot_doc, verbosity_level)

    prop.set("NAME", "lastUsedAnnotationId")
    prop.text = str(max_annot_idx)
    
    
    #Linguistic Type
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Defining LINGUISTIC_TYPEs")
    ling_type = ET.SubElement(annot_doc, "LINGUISTIC_TYPE")
    ling_type.set("LINGUISTIC_TYPE_ID", "base")
    ling_type.set("TIME_ALIGNABLE", "true")
    
    ling_type_par = ET.SubElement(annot_doc, "LINGUISTIC_TYPE")
    ling_type_par.set("LINGUISTIC_TYPE_ID", "parent")
    ling_type_par.set("CONSTRAINTS", "Included_In")
    ling_type_par.set("TIME_ALIGNABLE", "true")

    ling_type_par = ET.SubElement(annot_doc, "LINGUISTIC_TYPE")
    ling_type_par.set("LINGUISTIC_TYPE_ID", "symbolic")
    ling_type_par.set("CONSTRAINTS", "Symbolic_Association")
    ling_type_par.set("TIME_ALIGNABLE", "false")
    #Locale
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Setting LOCALE to '%s'." % lo.getdefaultlocale()[0])
    
    country_code = lo.getdefaultlocale()[0].split('_')[0]
    language_code = lo.getdefaultlocale()[0].split('_')[1]
    locale = ET.SubElement(annot_doc, "LOCALE")
    locale.set("COUNTRY_CODE", country_code)
    locale.set("LANGUAGE_CODE", language_code)

    #Pretty print the XML
    if verbosity_level >= 3:
        print("DEBUG: mausbpf2eaf : Formatting the XML to be human readable")
    indent(annot_doc)

    if verbosity_level >= 1:
        print("DEBUG: mausbpf2eaf : Writing XML to file %s" % out_file)
    tree = ET.ElementTree(annot_doc)
    tree.write(out_file, xml_declaration=True, encoding="utf-8", method="xml")
    

def indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def make_time_order(data, annot_doc, verbosity_level):
    """
    Constructs the time order from all existing class4 tracks.
    Every track is read and for every start and end point of an utterance a single node is created.
    The nodes are later matched to annotations by their index in make_tiers().
    """
    class4_list = []
    for time_tier in data["class4"]:
        if data["class4"][time_tier]:
            class4_list.append(time_tier)
    if not class4_list and not data["class2"]["SPD"]:
        sys.exit("ERROR: mausbpf2eaf : Could not find any time-alignable tracks in the input file. The .par input must at least contain one of the following tracks: %s - exiting" %", ".join(data["class4"]))
    
    elif class4_list == ['TRN']:
        sys.stderr.write("WARNING: mausbpf2eaf : The input file only contains a time-alignable TRN track which is not enough to build hierarchy information. Only writing out TRN data...")
    elif not class4_list and data["class2"]["SPD"]:
        sys.stderr.write("WARNING: mausbpf2eaf : The input file only contains a time-alignable SPD track which is not enough to build hierarchy information. Only writing out SPD data...")

    if verbosity_level >= 2:
        class4_list_output = ", ".join(class4_list)
        print("DEBUG: mausbpf2eaf : Constructing Time reference Levels from found class4 tiers: %s" %class4_list_output)

    time_order = ET.SubElement(annot_doc, "TIME_ORDER")
    sample_rate = int(data["header"]["SAM"])
    if not data["header"]["SAM"]:
        sys.exit("ERROR: mausbpf2eaf : Could not read sample rate from input file. Please make sure there is a valid value in SAM - exiting")

    all_times = []
    for track_name,time_values in data["class4"].items():
        if time_values:
            idx = 0
            for value in time_values:
                #Start
                try:
                    start_time = math.floor(int(value.split()[0])/sample_rate*1000)
                except ValueError:
                    sys.exit("ERROR: mausbpf2eaf : The input file seems to be malformed at value '%s' in track '%s'. Please reformat the input." % (value, track_name)) 
                time_id = track_name.lower() + str(idx)
                idx += 1
                if start_time < 0:
                    start_time = 0
                all_times.append((start_time,time_id))
                try:
                    end_time = math.floor((int(value.split()[0]) + int(value.split()[1]) + 1)/sample_rate*1000)
                except ValueError:
                    sys.exit("ERROR: mausbpf2eaf : The input file seems to be malformed at value '%s' in track '%s'. Please reformat the input." % (value, track_name)) 
                if end_time < 0:
                    end_time = 0
                time_id = track_name.lower() + str(idx)
                all_times.append((end_time,time_id))
                idx += 1
    
    for track_name,time_values in data["class2"].items():
        if time_values:
            idx = 0
            for value in time_values:
                #Start
                try:
                    start_time = math.floor(int(value.split()[0])/sample_rate*1000)
                except ValueError:
                    sys.exit("ERROR: mausbpf2eaf : The input file seems to be malformed at value '%s' in track '%s'. Please reformat the input." % (value, track_name)) 
                time_id = track_name.lower() + str(idx)
                idx += 1
                if start_time < 0:
                    sys.stderr.write("WARNING: mausbpf2eaf : The input file contains a class2 time value: '%d' in track '%s' which is smaller than 0. Setting value to zero...\n" % (start_time, track_name))
                    start_time = 0
                all_times.append((start_time,time_id))
                try:
                    end_time = math.floor((int(value.split()[0]) + int(value.split()[1]) + 1)/sample_rate*1000)
                except ValueError:
                    sys.exit("ERROR: mausbpf2eaf : The input file seems to be malformed at value '%s' in track '%s'. Please reformat the input." % (value, track_name)) 
                time_id = track_name.lower() + str(idx)
                if end_time < 0:
                    sys.stderr.write("WARNING: mausbpf2eaf : The input file contains a class2 time value: '%d' in track '%s' which is smaller than 0. Setting value to zero...\n" % (start_time, track_name))
                    end_time = 0
                all_times.append((end_time,time_id))
                idx += 1

    #Sort the time order entries by ascending order instead of all of one track and then all of the next track
    all_times.sort(key=lambda time_tuple: time_tuple[0])
    for time_tuple in all_times:
        time_slot = ET.SubElement(time_order, "TIME_SLOT")
        time_slot.set("TIME_SLOT_ID", str(time_tuple[1]))
        time_slot.set("TIME_VALUE", str(time_tuple[0]))

            
def get_time_reference_tier(data):
    """
    Defines the reference tier to be used for class1 and class1mult annotations, so they can be matched to specific times.
    The default is set to MAU.
    """
    time_reference_tier = None
    if data["class4"]["MAU"]:
        time_reference_tier = "MAU"
    elif data["class4"]["SAP"]:
        time_reference_tier = "SAP"
    elif data["class4"]["WOR"]:
        time_reference_tier = "WOR"
    elif data["class4"]["PHO"]:
        time_reference_tier = "PHO"
    elif data["class4"]["MAS"]:
        time_reference_tier = "MAS"
        
    return time_reference_tier

def get_symbolic_reference_tier(data):
    """
    Defines the reference tier to be used for class4 annotations. 
    Defines one class4 to be connected to the class1 reference tier.
    All others are linked to that "symbolic_reference_tier".
    """
    sym_reference_tier = None
    if data["class1"]["ORT"]:
        sym_reference_tier = "ORT"
    elif data["class1"]["KAN"]:
        sym_reference_tier = "KAN"
    elif data["class1"]["KSS"]:
        sym_reference_tier = "KSS"
    elif data["class1"]["KAS"]:
        sym_reference_tier = "KAS"
    elif data["class1"]["SPK"]:
        sym_reference_tier = "SPK"
    elif data["class1"]["MRP"]:
        sym_reference_tier = "MRP"
        
    return sym_reference_tier

def make_tiers(data, annot_doc, verbosity_level):
    """
    Constructs the annotation tiers and the containing annotations.
    The reference used is the word index in case of class1 and class1mult and the literal index (i.e line occurence) in case of class4.
    The index is then matched to a corresponding time order object from make_time_order().
    In case of class1 and class1mult the time points are computed by taking the word_number value and parsing through the class4 entries for that value, to find the start and end time.
    """
    time_reference_tier = get_time_reference_tier(data)
    sym_reference_tier = get_symbolic_reference_tier(data)
    annot_idx = 0
    
    class1_reference_map = []
    if time_reference_tier is not None:
        for track_class_name,track_data in data.items():
            if track_class_name is not "header":
                for track_name,track_items in track_data.items():
                    if track_items:
                        if verbosity_level >= 2:
                            print("DEBUG: mausbpf2eaf : Creating annotations for Tier: '%s'" % track_name)
                        tier = ET.SubElement(annot_doc, "TIER")
                        tier.set("ANNOTATOR", "MAUS")
                        tier.set("TIER_ID", track_name)
                        idx = 0
                        for track_value in track_items:
                            annotation = ET.SubElement(tier, "ANNOTATION")
                            if verbosity_level >= 3:
                                print("DEBUG: mausbpf2eaf : Creating annotation: '%s'" % track_value)
                            
                            if track_class_name == "class4":
                                align_annotation = ET.SubElement(annotation, "ALIGNABLE_ANNOTATION")
                                align_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                            
                                annotation_val = ET.SubElement(align_annotation, "ANNOTATION_VALUE")
                                
                                align_annotation.set("TIME_SLOT_REF1", track_name.lower() + str(idx))
                                align_annotation.set("TIME_SLOT_REF2", track_name.lower() + str(idx+1))
                                normalized_value_array = track_value.replace(" ","\t").split('\t')
                                annotation_val.text = " ".join(normalized_value_array[3:])

                                tier.set("LINGUISTIC_TYPE_REF", "base")
                                
                                idx += 2
                            
                            elif track_class_name == "class2":
                                align_annotation = ET.SubElement(annotation, "ALIGNABLE_ANNOTATION")
                                align_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                            
                                annotation_val = ET.SubElement(align_annotation, "ANNOTATION_VALUE")
                                
                                align_annotation.set("TIME_SLOT_REF1", track_name.lower() + str(idx))
                                align_annotation.set("TIME_SLOT_REF2", track_name.lower() + str(idx+1))
                                
                                #Characters after space->number->space->number->space
                                end_of_numbers = re.match("(\s+)?\d+\s+\d+\s+", track_value).end()
                                annotation_val.text = track_value[end_of_numbers:]
                                tier.set("LINGUISTIC_TYPE_REF", "base")
                                
                                idx += 2

                            elif track_class_name == "class1":
                                if track_name == sym_reference_tier:
                                    align_annotation = ET.SubElement(annotation, "ALIGNABLE_ANNOTATION")
                                    align_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                                
                                    annotation_val = ET.SubElement(align_annotation, "ANNOTATION_VALUE")
                                    
                                    start,end = get_time_info(str(idx), time_reference_tier, data)
                                    align_annotation.set("TIME_SLOT_REF1", time_reference_tier.lower() + str(start))
                                    align_annotation.set("TIME_SLOT_REF2", time_reference_tier.lower() + str(end))
                                    
                                    #Get first element after number and whitespace
                                    start_of_value = re.compile('(\s+)?\d+\s+').match(track_value).end()
                                    annotation_val.text = track_value[start_of_value:]
                                    
                                    tier.set("PARENT_REF", time_reference_tier)
                                    tier.set("LINGUISTIC_TYPE_REF", "parent")
                                    
                                    class1_reference_map.append((idx, annot_idx))
                                    
                                    idx += 1

                                    
                                else:
                                    ref_annotation = ET.SubElement(annotation, "REF_ANNOTATION")
                                    ref_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                                    
                                    wor_ref = [x[1] for x in class1_reference_map if x[0] == idx]
                                    if not wor_ref:
                                        sys.exit("ERROR: mausbpf2eaf : Class1 reference without corresponding number in the main tier: '%s' has been found" %(sym_reference_tier))
                                    ref_annotation.set("ANNOTATION_REF", "a" + str(wor_ref[0]))
                                    
                                    annotation_val = ET.SubElement(ref_annotation, "ANNOTATION_VALUE")
                                    start_of_value = re.compile('(\s+)?\d+\s+').match(track_value).end()
                                    annotation_val.text = track_value[start_of_value:]
                                    
                                    tier.set("PARENT_REF", sym_reference_tier)
                                    tier.set("LINGUISTIC_TYPE_REF", "symbolic")

                                    idx += 1
                            
                            elif track_class_name == "class1mult":
                                align_annotation = ET.SubElement(annotation, "ALIGNABLE_ANNOTATION")
                                align_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                            
                                annotation_val = ET.SubElement(align_annotation, "ANNOTATION_VALUE")
                                
                                time_reference = track_value.replace(" ","\t").split('\t')[0]
                                start,end = get_time_info_mult(time_reference, time_reference_tier, data)
                                align_annotation.set("TIME_SLOT_REF1", time_reference_tier.lower() + str(start))
                                align_annotation.set("TIME_SLOT_REF2", time_reference_tier.lower() + str(end))
                                
                                normalized_value_array = track_value.replace(" ","\t").split('\t')
                                annotation_val.text = " ".join(normalized_value_array[1:])
                                
                                tier.set("PARENT_REF", time_reference_tier)
                                tier.set("LINGUISTIC_TYPE_REF", "parent")
                                
                            annot_idx += 1
    else:
        if data["class4"]["TRN"]:
            tier = ET.SubElement(annot_doc, "TIER")
            tier.set("ANNOTATOR", "MAUS")
            tier.set("TIER_ID", "TRN")
            idx = 0
            for track_value in data["class4"]["TRN"]:
                annotation = ET.SubElement(tier, "ANNOTATION")
                if verbosity_level >= 3:
                    print("DEBUG: mausbpf2eaf : Creating annotation: '%s'" % track_value)
                
                align_annotation = ET.SubElement(annotation, "ALIGNABLE_ANNOTATION")
                align_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                annot_idx += 1
                
                annotation_val = ET.SubElement(align_annotation, "ANNOTATION_VALUE")
                
                align_annotation.set("TIME_SLOT_REF1", "trn" + str(idx))
                align_annotation.set("TIME_SLOT_REF2", "trn" + str(idx+1))
                
                normalized_value_array = track_value.replace(" ","\t").split('\t')
                annotation_val.text = " ".join(normalized_value_array[3:])

                tier.set("LINGUISTIC_TYPE_REF", "base")
                
                idx += 2
        if data["class2"]["SPD"]:
            tier = ET.SubElement(annot_doc, "TIER")
            tier.set("ANNOTATOR", "MAUS")
            tier.set("TIER_ID", "SPD")
            idx = 0
            for track_value in data["class2"]["SPD"]:
                annotation = ET.SubElement(tier, "ANNOTATION")
                if verbosity_level >= 3:
                    print("DEBUG: mausbpf2eaf : Creating annotation: '%s'" % track_value)
                align_annotation = ET.SubElement(annotation, "ALIGNABLE_ANNOTATION")
                align_annotation.set("ANNOTATION_ID", "a" +  str(annot_idx))
                annot_idx += 1
            
                annotation_val = ET.SubElement(align_annotation, "ANNOTATION_VALUE")
                
                align_annotation.set("TIME_SLOT_REF1", "spd" + str(idx))
                align_annotation.set("TIME_SLOT_REF2", "spd" + str(idx+1))
                
                #Characters after space->number->space->number->space
                end_of_numbers = re.match("(\s+)?\d+\s+\d+\s+", track_value).end()
                annotation_val.text = track_value[end_of_numbers:]
                tier.set("LINGUISTIC_TYPE_REF", "base")
                
                idx += 2

    return annot_idx-1

def get_time_info_mult(time_reference, time_reference_tier, data):
    start,end = 0,0
    if ';' in time_reference:
        first_word = time_reference.split(';')[0]
        second_word = time_reference.split(';')[1]
        wrong_start,start = get_time_info(first_word, time_reference_tier, data)
        end,wrong_end = get_time_info(second_word, time_reference_tier, data)
    elif ',' in time_reference:
        first_word = time_reference.split(',')[0]
        last_word = time_reference.split(',')[-1]
        start,wrong_end = get_time_info(first_word, time_reference_tier, data)
        wrong_start,end = get_time_info(last_word, time_reference_tier, data)
    else:
        start,end = get_time_info(time_reference, time_reference_tier, data)
    return start,end

def get_time_info(word_index, time_reference_tier, data):
    """
    It's always index *2 because ELAN does not allow for the same time order element to be used 
    by different annotations without hierarchy reference. That's why it was decided
    to make a time reference element for every start and end point even though that duplicates entries.
    """
    possible_values = []
    for time_stamp in data["class4"][time_reference_tier]:
        time_stamp_class1_ref = int(time_stamp.replace(" ","\t").split('\t')[2])
        possible_values.append(time_stamp_class1_ref)

    if int(word_index) not in possible_values:
        sys.stderr.write("WARNING: mausbpf2eaf : The input file contains a class1 reference: %d ,that does not occur in time reference track: '%s'.\n" % (int(word_index), time_reference_tier))

    first_index = 0
    last_index = 0
    for value in data["class4"][time_reference_tier]:
        class4_kan_ref = value.replace(" ","\t").split('\t')[2]
        if class4_kan_ref == word_index and first_index == 0:
            first_index = data["class4"][time_reference_tier].index(value) * 2
            last_index = data["class4"][time_reference_tier].index(value) * 2 + 1
        elif class4_kan_ref == word_index and first_index != 0:
            last_index = data["class4"][time_reference_tier].index(value) * 2 + 1
  
    
    return first_index,last_index

    

class ParParser: 
    """
    Parses Partitur file and writes all supported tracks in the dictionary below.
    Class1 is the only one that is in order, because there it is necessary to remember the annotation index in make_tiers().
    Only header information currently in use is SAM, as the samplerate is needed for conversion.
    """
    data = OrderedDict([("header", {"SAM": 0}),
           ("class1", OrderedDict([("ORT", []), ("KAN", []), ("KSS", []), ("KAS", []), ("SPK", []), ("MRP", [])])),
           ("class1mult", OrderedDict([("NOI", []), ("TRL", []), ("TR2", []), ("TRO", [])])),
           ("class2", OrderedDict([("SPD", [])])),
           ("class4", OrderedDict([("MAU", []), ("MAS", []), ("PHO", []), ("WOR", []), ("SAP", []), ("TRN", [])]))])


    def __init__(self, filepath):
        self.source_file = os.path.abspath(filepath)
   

    def read_file(self, verbosity_level):
        
        if verbosity_level >= 1:
            print("DEBUG: mausbpf2eaf : Parsing partitur file '%s'" % self.source_file)
        
        with open(self.source_file, 'r') as source_data:
            lines = source_data.readlines()
            for line in lines:
                for track_class in self.data:
                    if track_class == "header":
                        for track_name in self.data[track_class]:
                            if line.startswith(track_name):
                                self.data[track_class][track_name] = line.split()[1]
                    else:
                        for track_name in self.data[track_class]:
                            if line.startswith(track_name):
                                self.data[track_class][track_name].append(line.rstrip()[5:])
        
        if verbosity_level >= 2:
            found_tracks = []
            for track_class in self.data:
                for track in self.data[track_class]:
                    if self.data[track_class][track]:
                        found_tracks.append(track)
            print("DEBUG: mausbpf2eaf : Found Tracks: '%s'" % ", ".join(found_tracks))
            if 'class1mult' and not 'class1':
                sys.exit("ERROR: mausbpf2eaf : Found a class1 track with multiple references but no plain class1 track that can serve as reference level - exiting")
        return(self.data)

def check_output_validity(out_file, verbosity_level):
    if verbosity_level >= 1:
        print("DEBUG: mausbpf2eaf : Checking validity of constructed eaf file.")
    xsd_path = os.path.dirname(os.path.abspath(__file__)) + '/EAFv3.0.xsd'
    xmllint_process = subprocess.run(['xmllint', '--noout', '--schema', xsd_path, out_file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if xmllint_process.returncode != 0:
        os.remove(out_file)
        sys.exit("ERROR: mausbpf2eaf : Xmllint stopped with exit status '%d' indicating that the constructed eaf file is malformed - exiting" % xmllint_process.returncode)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description="Parse Input Arguments.")
    argument_parser.add_argument("--INP", metavar="FILE",
                                help="Input BPF .par file to be converted")
    
    argument_parser.add_argument("--OUT", metavar="FILE", help="Output path where the final .eaf file will be stored")
    
    argument_parser.add_argument("--verbose", "-v", default=0, help="Verbosity level", action="count")
    argument_parser.add_argument("--version", help="Outputs version number", action="store_true")
    
    args = argument_parser.parse_args()

    version_number= "1.5"
    if args.version:
        print(version_number)
    else:
        if not args.INP:
            sys.exit("ERROR: mausbpf2eaf : Missing required input argument --INP. Please specify a valid BAS partitur file to be converted.")
        is_valid_par_file(args.INP)
        #valid_audio_ext = ["wav", "mp3", "flac"]
        orig_wav_path = os.path.splitext(os.path.basename(args.INP))[0] + ".wav"
        
        out_file = os.path.splitext(orig_wav_path)[0] + ".eaf"
        if args.OUT:
            if is_valid_output_file(args.OUT):
                out_file = args.OUT
        else:
            sys.stderr.write("WARNING: mausbpf2eaf : No output file specified. Writing eaf to default location: '%s'\n" %os.path.abspath(out_file))

        
        parParser = ParParser(args.INP)
        data = parParser.read_file(args.verbose)
        
        write_elan_data(data, out_file, orig_wav_path, args.verbose)
        
        check_output_validity(out_file, args.verbose)
        

